#include <stdio.h> // printf
#include <stdlib.h> // calloc
#include <assert.h> // assert
#include <cuda.h> // CUDA library
#include <cutil.h> // CUDA error checking

// integer ceiling division
#define DIV_CEIL(a,b) (((a)+(b)-1)/(b))
// maximum
#define MIN(a,b) (((a)<(b))?(a):(b))

// beware, much less than 65535, but it is power of two.
#define MAX_GRID_SIZE 32768

#define BLOCK_SIZE 256
#define NUM_THREADS BLOCK_SIZE
#define WARP_SIZE 32
#define NUM_WARPS (NUM_THREADS/WARP_SIZE)


__global__ void set_values(const uint N, int *values) {
	int tid = threadIdx.x;
	// 2d <-> 1d transformation to overcome 16-bit limit on grid size.
	int bid = blockIdx.y*gridDim.x + blockIdx.x;
	int i = bid*blockDim.y*blockDim.x + tid;
	if (i<N)
		values[i] = (int) WARP_SIZE*cosf((float)i);
}

// To avoid bank conflicts, or not!
#define WARP_STRIDE (WARP_SIZE)

__global__ void multi_scan(const int *values, int *inclusive, int *exclusive) {
	__shared__ volatile int scan[NUM_WARPS*WARP_STRIDE];
	int tid = threadIdx.x;
	int warp = tid/WARP_SIZE;
	int lane = (WARP_SIZE-1) & tid;
	// 2d <-> 1d transformation to overcome 16-bit limit on grid size.
	int bid = blockIdx.y*gridDim.x + blockIdx.x;
	int i = bid*blockDim.y*blockDim.x + tid;

	// Read from global memory.
	int x = values[i];
	int s = warp*WARP_STRIDE+lane;
	scan[s] = x;

	// Run inclusive scan on each warp's data.
	int sum = x;
	#pragma unroll
	for (int offset=1; offset<WARP_SIZE; offset*=2) {
		if (lane>=offset) sum += scan[s-offset];
		scan[s] = sum;
	}

	// Synchronize to make all the totals available to the reduction code.
	__syncthreads();

	// Reducition of the totals
	__shared__ volatile int totals[NUM_WARPS];
	if (tid < NUM_WARPS) {
		// Grab the block total for the tid'th block. This is the last element
		// in the block's scanned sequence.
		int total = scan[tid*WARP_STRIDE + WARP_SIZE-1];
		totals[tid] = total;

		int totalsSum = total;
		#pragma unroll
		for(int offset=1; offset<NUM_WARPS; offset*=2) {
			if (tid>=offset) totalsSum += totals[tid-offset];
			totals[tid] = totalsSum;
		}

		// Subtract total from totalsSum for an exclusive scan.
		totals[tid] = totalsSum - total;
	}

	// Synchronize to make the block scan available to all warps.
	__syncthreads();

	// Add the block scan to the inclusive sum for the block.
	sum += totals[warp];

	// Write the inclusive and exclusive scans to global memory.
	inclusive[i] = sum;
	exclusive[i] = sum - x;
}


int main(int argc, char **argv)
{
	int *d_values = NULL, *h_values = NULL; // pointers in Host and Device
	int *d_inclusive = NULL, *h_inclusive = NULL;
	int *d_exclusive = NULL, *h_exclusive = NULL;
	assert(argc==2);
	const unsigned int N = 1<<atoi(argv[1]); // vector size
	assert(N%WARP_SIZE==0);

	CUDA_SAFE_CALL(cudaMalloc(&d_values, N * sizeof(int)));
	CUDA_SAFE_CALL(cudaMalloc(&d_inclusive, N * sizeof(int)));
	CUDA_SAFE_CALL(cudaMalloc(&d_exclusive, N * sizeof(int)));
	h_values = (int *) calloc(N, sizeof(int));
	h_inclusive = (int *) calloc(N, sizeof(int));
	h_exclusive = (int *) calloc(N, sizeof(int));
	assert(d_values && h_values && d_inclusive && h_inclusive && d_exclusive && h_exclusive);

	dim3 block_size(MIN(BLOCK_SIZE, N));
	unsigned int blocks = DIV_CEIL(N,BLOCK_SIZE);
	dim3 grid_size(MIN(MAX_GRID_SIZE,blocks), DIV_CEIL(blocks,MAX_GRID_SIZE));
	//printf("%d %d %d \n", block_size.x, grid_size.x, grid_size.y);
	set_values<<<grid_size, block_size>>>(N, d_values);
	CUT_CHECK_ERROR("set_values kernel failed");
	multi_scan<<<grid_size, block_size>>>((const int*)d_values, d_inclusive, d_exclusive);
	CUT_CHECK_ERROR("multi_scan kernel failed");
	cudaDeviceSynchronize();

	CUDA_SAFE_CALL(cudaMemcpy(h_values, d_values, N*sizeof(int), cudaMemcpyDefault));
	CUDA_SAFE_CALL(cudaMemcpy(h_inclusive, d_inclusive, N*sizeof(int), cudaMemcpyDefault));
	CUDA_SAFE_CALL(cudaMemcpy(h_exclusive, d_exclusive, N*sizeof(int), cudaMemcpyDefault));

	unsigned int block=0;
	unsigned int thread=0;
	for (block=0; block<N/BLOCK_SIZE; block++) {
		int inclusive = 0, exclusive = 0;
		for (thread=0; thread<BLOCK_SIZE; ++thread) {
			unsigned int i = block*BLOCK_SIZE + thread;
			inclusive += h_values[i];
			if (inclusive!=h_inclusive[i] || exclusive!=h_exclusive[i]) {
				goto EXIT;
			}
			exclusive += h_values[i];
		}
	}
EXIT:
	if (N/BLOCK_SIZE!=block) { // if error, print complete block
		printf("Error in block %d, thread %d\n", block, thread);
		for (unsigned int t=0; t<BLOCK_SIZE; ++t) {
			printf("%d ", h_values[block+t]);
		} printf("\n"); printf("--------\n");
		for (unsigned int t=0; t<BLOCK_SIZE; ++t) {
			printf("%d ", h_inclusive[block+t]);
		} printf("\n"); printf("--------\n");
		for (unsigned int t=0; t<BLOCK_SIZE; ++t) {
			printf("%d ", h_exclusive[block+t]);
		} printf("\n"); printf("--------\n");
	}

	free(h_values); free(h_inclusive); free(h_exclusive);
	CUDA_SAFE_CALL(cudaFree(d_values));
	CUDA_SAFE_CALL(cudaFree(d_inclusive));
	CUDA_SAFE_CALL(cudaFree(d_exclusive));

	return 0;
}
