#include <stdio.h> // printf
#include <stdlib.h> // calloc
#include <assert.h> // assert
#include <cuda.h> // CUDA library
#include <cutil.h> // CUDA error checking

// integer ceiling division
#define DIV_CEIL(a,b) (((a)+(b)-1)/(b))
// maximum
#define MIN(a,b) (((a)<(b))?(a):(b))

// beware, much less than 65535, but it is power of two.
#define MAX_GRID_SIZE 32768

#define BLOCK_SIZE 256
#define NUM_THREADS BLOCK_SIZE
#define WARP_SIZE 32
#define LOG_WARP_SIZE 5
#define NUM_WARPS (NUM_THREADS/WARP_SIZE)


__global__ void set_values(const uint N, int *values) {
	int tid = threadIdx.x;
	// 2d <-> 1d transformation to overcome 16-bit limit on grid size.
	int bid = blockIdx.y*gridDim.x + blockIdx.x;
	int i = bid*blockDim.y*blockDim.x + tid;
	if (i<N)
		values[i] = (int) WARP_SIZE*cosf((float)i);
}


__global__ void warp_scan1(const int *values, int *inclusive, int *exclusive) {
	__shared__ volatile int scan[NUM_WARPS*WARP_SIZE];
	int tid = threadIdx.x;
	int warp = tid/WARP_SIZE;
	int lane = (WARP_SIZE-1) & tid;
	// 2d <-> 1d transformation to overcome 16-bit limit on grid size.
	int bid = blockIdx.y*gridDim.x + blockIdx.x;
	int i = bid*blockDim.y*blockDim.x + tid;

	// read from global
	int x = values[i];
	scan[tid] = x;

	int sum = x;
	for (int offset=1; offset<WARP_SIZE; offset*=2) {
		if (lane>=offset) sum += scan[tid-offset];
		scan[tid] = sum;
	}
	inclusive[i] = sum;
	exclusive[i] = sum - x;
}

#define SCAN_STRIDE (WARP_SIZE + WARP_SIZE/2 + 1)

__global__ void warp_scan2(const int *values, int *inclusive, int *exclusive) {
	// Reserve a half warp of extra space.
	__shared__ volatile int scan[NUM_WARPS*SCAN_STRIDE];
	int tid = threadIdx.x;
	int warp = tid/WARP_SIZE;
	int lane = (WARP_SIZE-1) & tid;
	// 2d <-> 1d transformation to overcome 16-bit limit on grid size.
	int bid = blockIdx.y*gridDim.x + blockIdx.x;
	int i = bid*blockDim.y*blockDim.x + tid;

	// Zero out the values at the start of the scan array and make a pointer
	// to scan + WARP_SIZE / 2. Now s is addressable and zero for the first
	// sixteen values before the start of the array. This eliminates comparisons
	// and predicated execution.
	volatile int* s = scan + SCAN_STRIDE*warp + lane + WARP_SIZE/2;
	s[-(WARP_SIZE/2)] = 0;

	// Read from global memory.
	int x = values[i];
	s[0] = x;

	// Run each pass of the scan.
	int sum = x;
	for (int offset=1; offset<WARP_SIZE; offset*=2) {
		// Since offset is at most 16, -offset will not read past the beginning
		// of the scan array.
		sum += s[-offset];
		s[0] = sum;
	}

	// Write sum to inclusive and sum - x (the original source element) to
	// exclusive.
	inclusive[i] = sum;
	exclusive[i] = sum - x;
}


int main(int argc, char **argv)
{
	int *d_values = NULL, *h_values = NULL; // pointers in Host and Device
	int *d_inclusive = NULL, *h_inclusive = NULL;
	int *d_exclusive = NULL, *h_exclusive = NULL;
	assert(argc==2);
	const unsigned int N = 1<<atoi(argv[1]); // vector size
	assert(N%WARP_SIZE==0);

	CUDA_SAFE_CALL(cudaMalloc(&d_values, N * sizeof(int)));
	CUDA_SAFE_CALL(cudaMalloc(&d_inclusive, N * sizeof(int)));
	CUDA_SAFE_CALL(cudaMalloc(&d_exclusive, N * sizeof(int)));
	h_values = (int *) calloc(N, sizeof(int));
	h_inclusive = (int *) calloc(N, sizeof(int));
	h_exclusive = (int *) calloc(N, sizeof(int));
	assert(d_values && h_values && d_inclusive && h_inclusive && d_exclusive && h_exclusive);

	dim3 block_size(MIN(BLOCK_SIZE, N));
	unsigned int blocks = DIV_CEIL(N,BLOCK_SIZE);
	dim3 grid_size(MIN(MAX_GRID_SIZE,blocks), DIV_CEIL(blocks,MAX_GRID_SIZE));
	//printf("%d %d %d \n", block_size.x, grid_size.x, grid_size.y);
	set_values<<<grid_size, block_size>>>(N, d_values);
	CUT_CHECK_ERROR("set_values kernel failed");
	warp_scan1<<<grid_size, block_size>>>((const int*)d_values, d_inclusive, d_exclusive);
	CUT_CHECK_ERROR("warp_scan1 kernel failed");
	cudaDeviceSynchronize();

	CUDA_SAFE_CALL(cudaMemcpy(h_values, d_values, N*sizeof(int), cudaMemcpyDefault));
	CUDA_SAFE_CALL(cudaMemcpy(h_inclusive, d_inclusive, N*sizeof(int), cudaMemcpyDefault));
	CUDA_SAFE_CALL(cudaMemcpy(h_exclusive, d_exclusive, N*sizeof(int), cudaMemcpyDefault));

	unsigned int i=0;
	unsigned int lane=0;
	for (i=0; i<N; i+=WARP_SIZE) {
		int inclusive = 0, exclusive = 0;
		for (lane=0; lane<WARP_SIZE; ++lane) {
			inclusive += h_values[i+lane];
			if (inclusive!=h_inclusive[i+lane] || exclusive!=h_exclusive[i+lane]) {
				goto EXIT;
			}
			exclusive += h_values[i+lane];
		}
	}
EXIT:
	if (N!=i) { // if error, print complete lane.
		printf("Error in warp %d, lane %d\n", i/WARP_SIZE, lane);
		for (unsigned int lane=0; lane<WARP_SIZE; ++lane) {
			printf("%d ", h_values[i+lane]);
		} printf("\n");
		for (unsigned int lane=0; lane<WARP_SIZE; ++lane) {
			printf("%d ", h_inclusive[i+lane]);
		} printf("\n");
		for (unsigned int lane=0; lane<WARP_SIZE; ++lane) {
			printf("%d ", h_exclusive[i+lane]);
		} printf("\n");
	}

	free(h_values); free(h_inclusive); free(h_exclusive);
	CUDA_SAFE_CALL(cudaFree(d_values));
	CUDA_SAFE_CALL(cudaFree(d_inclusive));
	CUDA_SAFE_CALL(cudaFree(d_exclusive));

	return 0;
}
